/**
 * Copyright (c) 2014 - 2019, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/** @file
 *
 * @defgroup ble_sdk_uart_over_ble_main main.c
 * @{
 * @ingroup  ble_sdk_app_nus_eval
 * @brief    UART over BLE application main file.
 *
 * This file contains the source code for a sample application that uses the Nordic UART service.
 * This application uses the @ref srvlib_conn_params module.
 */


#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf.h"
#include "ble_hci.h"
#include "ble_advdata.h"
#include "ble_advertising.h"
#include "ble_conn_params.h"
#include "nrf_sdh.h"
#include "nrf_sdh_soc.h"
#include "nrf_sdh_ble.h"
#include "nrf_ble_gatt.h"
#include "nrf_ble_qwr.h"
#include "ble_dis.h"
#include "app_timer.h"
#include "ble_nus.h"
#include "app_uart.h"
#include "app_util_platform.h"
#include "bsp_btn_ble.h"
#include "nrf_pwr_mgmt.h"

#include "nrf_drv_clock.h"
#include "app_scheduler.h"
#include "mem_manager.h"

#include "SEGGER_RTT.h"

#define DEVICE_NAME                     "xTag"                                  // As Friendlyname and other BLE services.
#define DEVICE_NAME_SUFFIX_LEN          14                                      // Extra space after for DEVICE_NAME + " xxxxxxxxxxxx" with \0 at end.
#define MANUFACTURER_NAME               "Deviceworx"                            // For BLE services.
#define APP_ADV_INTERVAL                800                                     // Mult of 0.625 ms. Use 1600 for 1 sec.
#define APP_BLE_OBSERVER_PRIO           3                                       // Default
#define APP_BLE_CONN_CFG_TAG            1                                       // SoftDevice BLE configuration.

#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(80, UNIT_1_25_MS)             /**< Minimum acceptable connection interval (20 ms), Connection interval uses 1.25 ms units. */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(80, UNIT_1_25_MS)             /**< Maximum acceptable connection interval (75 ms), Connection interval uses 1.25 ms units. */
#define SLAVE_LATENCY                   0                                           /**< Slave latency. */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(4000, UNIT_10_MS)             /**< Connection supervisory timeout (4 seconds), Supervision Timeout uses 10 ms units. */
#define FIRST_CONN_PARAMS_UPDATE_DELAY  APP_TIMER_TICKS(5000)                       /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (5 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY   APP_TIMER_TICKS(30000)                      /**< Time between each call to sd_ble_gap_conn_param_update after the first call (30 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT    3                                           /**< Number of attempts before giving up the connection parameter negotiation. */

#define DEAD_BEEF                       0xDEADBEEF                                  /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

#define BLE_MAX_TX_SVC_BYTES                244
#define BLE_MAX_RX_SVC_BYTES                8


// GUID gen for xTag online - 10c66c52-e5c9-4fe7-8a04-e40b13eb932c
#define BLE_UUID_XTAG_BASE_UUID         {0x10, 0xC6, 0x6C, 0x52, 0xE5, 0xC9, 0x4F, 0xE7, 0x8A, 0x04, 0xE4, 0x0B, 0x13, 0xEB, 0x93, 0x2C}
#define BLE_UUID_XTAG_SERVICE           0x0001 // 

#define BLE_UUID_XTAG_RX_CHARACTERISTC_UUID          0x0002
#define BLE_UUID_XTAG_TX_CHARACTERISTC_UUID          0x0003

// Scheduler settings
#define SCHED_MAX_EVENT_DATA_SIZE   sizeof(ble_nus_evt_t)
#define SCHED_QUEUE_SIZE            5

#define MFR_PAYLOAD_SIZE_GAP  14

typedef struct
{
    uint16_t conn_handle;                     // Handle of the current connection (as provided by the BLE stack, is BLE_CONN_HANDLE_INVALID if not in a connection).
    uint16_t service_handle;                  // Handle of Our Service (as provided by the BLE stack). 
    ble_gatts_char_handles_t rx_char_handles; // Hnndles for the rx char.
    ble_gatts_char_handles_t tx_char_handles; // Hnndles for the tx char.
}ble_os_t;

NRF_BLE_GATT_DEF(m_gatt);                                                 // GATT module instance.

//static uint16_t m_conn_handle = BLE_CONN_HANDLE_INVALID;                  // Handle of the current connection..
ble_os_t m_xtag_service;         

static bool m_notif_busy = false;     // Notification out when true - must wait for client confirmation.

// UUIDs for standard services here.
static ble_uuid_t m_adv_uuids[] =                       
{
  {BLE_UUID_DEVICE_INFORMATION_SERVICE, BLE_UUID_TYPE_BLE},
  {BLE_UUID_XTAG_SERVICE, BLE_UUID_TYPE_BLE }
};

static uint8_t rx_char_data[BLE_MAX_RX_SVC_BYTES];

ble_gap_adv_params_t m_adv_params;                                        // Parameters to be passed to the stack when starting advertising.
uint8_t              m_adv_handle = BLE_GAP_ADV_SET_HANDLE_NOT_SET;       // Advertising handle used to identify an advertising set.
uint8_t              m_enc_advdata1[BLE_GAP_ADV_SET_DATA_SIZE_MAX];       // 31 Byte Buffer for storing an encoded advertising set.
uint8_t              m_enc_advdata2[BLE_GAP_ADV_SET_DATA_SIZE_MAX];       // 31 Byte Buffer for storing an encoded advertising set.
uint8_t              m_enc_scanrespdata1[BLE_GAP_ADV_SET_DATA_SIZE_MAX];  // Buffer for storing an encoded advertising set.
uint8_t              m_enc_scanrespdata2[BLE_GAP_ADV_SET_DATA_SIZE_MAX];  // Buffer for storing an encoded advertising set.


// Setup 2 buffers for advertising and switching on the fly.
// Ref here: https://devzone.nordicsemi.com/f/nordic-q-a/32824/dynamically-updating-advertising-data-in-sdk-15
uint8_t m_bufIdx = 0;
ble_gap_adv_data_t m_gapData[2] = 
{
  { 
    .adv_data.p_data = m_enc_advdata1,
    .adv_data.len = sizeof(m_enc_advdata1),
    .scan_rsp_data.p_data = m_enc_scanrespdata1,
    .scan_rsp_data.len = sizeof(m_enc_scanrespdata1)
  },
  { 
    .adv_data.p_data = m_enc_advdata2,
    .adv_data.len = sizeof(m_enc_advdata2),
    .scan_rsp_data.p_data = m_enc_scanrespdata2,
    .scan_rsp_data.len = sizeof(m_enc_scanrespdata2)
  },
};                                         // As per egs, we need a svc struct.

APP_TIMER_DEF(m_100_msec_timer_id);                         // Misc stuff every 100 mSec.
static uint16_t m_100_msec_60sec_cnt = 0;                   // Used to inc the min ctr.
static uint16_t m_100_msec_10sec_cnt = 0;                   // Used to restart adverts.
APP_TIMER_DEF(m_25_msec_timer_id);                          // Acq data collected every 15 mSec.

static uint8_t* m_p_ramstore = NULL;    // Datastore for fake 240 bytes. Store values of 0 - 240 (0xF0) and rotate this data with each send.

static uint8_t m_pendingNotifications = 0;
static uint16_t m_handlerCount = 0;

void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name) { app_error_handler(DEAD_BEEF, line_num, p_file_name); }

void set_notif_busy(bool _busy) { m_notif_busy = _busy; }
bool is_notif_busy(void) { return m_notif_busy; }

void on_main_shed_rx_char_evt_handler(void *p_event_data, uint16_t event_size)
{
  SEGGER_RTT_printf(0, "Incomming bytes length %d - %02X %02X ... Starting notifications.\n", event_size, ((uint8_t*) p_event_data)[0], ((uint8_t*) p_event_data)[1]);

  // Nothing too fancy. Anything in - start sending notifications to the central from a timer handler.
  // 25 mSec period is how often we send 40 samples at a rate of 1600 per sec.
  APP_ERROR_CHECK(app_timer_start(m_25_msec_timer_id, APP_TIMER_TICKS(25), NULL));

  // We will only handle x notifications.
  m_handlerCount = 0;
}

void on_main_shed_data_notification_handler(void *p_event_data, uint16_t event_size)
{
  uint32_t err_code;
  uint8_t tempStore = 0;
  uint16_t pcktLen = 240;
  ble_gatts_hvx_params_t hvx_params;

  // Only process 400 times.
  if(m_handlerCount++ > 400)
  {
    SEGGER_RTT_printf(0, "Stopping notifications.\n"); 

    APP_ERROR_CHECK(app_timer_stop(m_25_msec_timer_id));

    return;
  }

  // New notification required whenever this handler is called.
  m_pendingNotifications++;

  // If we are still busy sending the last packet, cannot continue. m_pendingNotifications won't get reduced.
  // See BLE_GATTS_EVT_HVN_TX_COMPLETE event handling where notification busy flag is cleared.
  if(is_notif_busy())
  {
    return;
  }

  // Loop through pending notifications. We can send more than 1 every 15 mSec if deta length extensions are working.
  while((m_pendingNotifications>0) && (is_notif_busy() == false))
  {
    // Send the data.
    memset(&hvx_params, 0, sizeof(hvx_params));
    hvx_params.handle = m_xtag_service.tx_char_handles.value_handle;
    hvx_params.type   = BLE_GATT_HVX_NOTIFICATION;
    hvx_params.offset = 0;
    hvx_params.p_len  = &pcktLen;
    hvx_params.p_data = &(m_p_ramstore[0]);  

    err_code = sd_ble_gatts_hvx(m_xtag_service.conn_handle, &hvx_params);                       
    if ( (err_code == NRF_ERROR_RESOURCES) || (err_code == NRF_ERROR_BUSY) ) // Notification pending.
    {
      // Indicate that we are busy until a BLE_GATTS_EVT_HVN_TX_COMPLETE comes in.
      set_notif_busy(true);

      // DEBUG msgs.
      if(err_code == NRF_ERROR_BUSY) { SEGGER_RTT_printf(0, "Notification busy\n"); }
      else { SEGGER_RTT_printf(0, "Notification resource busy\n"); }
    }
    else if(err_code != NRF_SUCCESS)
    {
        SEGGER_RTT_printf(0, "Notification error: %d\n", err_code);

        // Some errors handled externally.
        return;
    }

    // Reduce m_pendingNotifications as we just handled one.
    m_pendingNotifications--;

    // Rotate the data by shuffling it up by 1 byte (Data at Idx 1 to Idx 0, etc). Data at Idx 0 to Idx 239.
    uint8_t tempStore = m_p_ramstore[0];
    for(int x=0; x<239; x++) { m_p_ramstore[x] = m_p_ramstore[x+1]; }
    m_p_ramstore[239] = tempStore;

  } // while()
}

static void timer_handler_100_msec(void* p_context)
{
  if(++m_100_msec_10sec_cnt >= 100)
  {
    // Restart advertising.
    sd_ble_gap_adv_start(m_adv_handle, APP_BLE_CONN_CFG_TAG);

    // Reset every cycle.
    m_100_msec_10sec_cnt = 0;
  }

  // Inc the min counter and handle periodic stuff.
  if(++m_100_msec_60sec_cnt >= 600)
  { 
    // Reset every min.
    m_100_msec_60sec_cnt = 0;

    SEGGER_RTT_printf(0, "Minute with mSec count %d.\n", (app_timer_cnt_get()/33));
  }

  return;
}

// Handler that emulates acquired data from a sensor. 25 samples (6 bytes each) every 15 mSec closely approx 1600 samples/sec/
static void timer_handler_25_msec(void* p_context)
{
  // Get out of handler quick. Schedule handler that runs on main thread.
  app_sched_event_put(NULL,  0, on_main_shed_data_notification_handler);

  return;
}

/**@brief Function for initializing the timer module.
 */
static void timers_init(void)
{
  // Initialize timer module.
  APP_ERROR_CHECK(app_timer_init());

  // Create the timer.
  APP_ERROR_CHECK(app_timer_create(&m_100_msec_timer_id, APP_TIMER_MODE_REPEATED, timer_handler_100_msec));

  APP_ERROR_CHECK(app_timer_create(&m_25_msec_timer_id, APP_TIMER_MODE_REPEATED, timer_handler_25_msec));
}

// Handle BLE events.
static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context)
{
  ret_code_t err_code = NRF_SUCCESS;

  ble_gap_phys_t phys;
  ble_gap_evt_phy_update_t const* p_phy_evt;
  ble_gap_evt_data_length_update_t const* p_datalen_evt;

  // SEGGER_RTT_printf(0, "Bluetooth Evt 0x%02X.\n", p_ble_evt->header.evt_id);

  switch (p_ble_evt->header.evt_id)
  {
    case BLE_GAP_EVT_ADV_SET_TERMINATED:

      //SEGGER_RTT_printf(0, "Adv Terminated.\n");

      break;

    case BLE_GAP_EVT_DISCONNECTED:

      SEGGER_RTT_printf(0, "Disconnect Evt.\n");

      m_xtag_service.conn_handle = BLE_CONN_HANDLE_INVALID;

      // xtag_set_device_state(DEV_STATE_DISCONN); // Update state.

      break;

    case BLE_GAP_EVT_CONNECTED:

      //NRF_LOG_INFO("Connected.");
      SEGGER_RTT_printf(0, "Connect Evt.\n");
      m_xtag_service.conn_handle = p_ble_evt->evt.gap_evt.conn_handle;

      // Log the effective mtu.
      SEGGER_RTT_printf(0, "Effective MTU after connection is: %d bytes\n", 
        nrf_ble_gatt_eff_mtu_get(&m_gatt, p_ble_evt->evt.gap_evt.conn_handle));

      // Required to work around nordic issue wherin adv stops upon connection. Suggested by Nordic.
      sd_ble_gap_adv_start(m_adv_handle, APP_BLE_CONN_CFG_TAG);

      break;

    case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
    
      SEGGER_RTT_printf(0, "Phy update request from central rx %d tx %d.\n", 
        p_ble_evt->evt.gap_evt.params.phy_update_request.peer_preferred_phys.rx_phys,
        p_ble_evt->evt.gap_evt.params.phy_update_request.peer_preferred_phys.tx_phys);

      // Handle the request that has been initiated in the server and accept the requested rx and tx phys.
      err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &p_ble_evt->evt.gap_evt.params.phy_update_request.peer_preferred_phys);
      APP_ERROR_CHECK(err_code);

      break;

    case BLE_GAP_EVT_PHY_UPDATE: // Log only for debugging.

      p_phy_evt = &p_ble_evt->evt.gap_evt.params.phy_update;

      if (p_phy_evt->status == BLE_HCI_STATUS_CODE_LMP_ERROR_TRANSACTION_COLLISION)
      {
        SEGGER_RTT_printf(0, "Phy update: Trans collision during updated.\n");
        break;
      }

      if(p_phy_evt->status == BLE_HCI_STATUS_CODE_SUCCESS)
      {
        SEGGER_RTT_printf(0, "Phy updated: Tx phy: 0x%02X Rx phy: 0x%02X.\n", p_phy_evt->tx_phy, p_phy_evt->rx_phy);
      }
      else
      {
        if(p_phy_evt->status == BLE_HCI_DIFFERENT_TRANSACTION_COLLISION) // Common.
        {
          SEGGER_RTT_printf(0, "Phy update failed due to transaction collision.\n");
        }
        else { SEGGER_RTT_printf(0, "Phy update failed. Status: 0x%02X.\n", p_phy_evt->status); }
      }

      break;

    case BLE_GAP_EVT_DATA_LENGTH_UPDATE:

      p_datalen_evt = &p_ble_evt->evt.gap_evt.params.data_length_update;

      SEGGER_RTT_printf(0, "Data length (PDU) update. Max Tx octets: %d, Max Rx octets %d, Max Tx uSec: %d, Max Rx uSec: %d\n",
        p_datalen_evt->effective_params.max_tx_octets, p_datalen_evt->effective_params.max_rx_octets, 
        p_datalen_evt->effective_params.max_tx_time_us, p_datalen_evt->effective_params.max_rx_time_us);

    case BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST:

      SEGGER_RTT_printf(0, "GATT MTU Request In.\n");

      break;

    case BLE_GATTS_EVT_HVN_TX_COMPLETE:

      // Ensure that our notification busy status is updated to indicate idle and another notification is ok.
      set_notif_busy(false);

      // DEBUG
      //SEGGER_RTT_printf(0, "Notification complete.\n");
      
      break;

    case BLE_GATTC_EVT_TIMEOUT:
      // Disconnect on GATT Client timeout event.
      SEGGER_RTT_printf(0, "GATT Client Timeout.");
      err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                       BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
      APP_ERROR_CHECK(err_code);
      break;

    case BLE_GATTS_EVT_TIMEOUT:
      // Disconnect on GATT Server timeout event.
      SEGGER_RTT_printf(0, "GATT Server Timeout.");
      err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                       BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
      APP_ERROR_CHECK(err_code);
      break;

    case BLE_GATTS_EVT_WRITE:
    {
      // Incoming data over the rx char.
       ble_gatts_evt_write_t const * p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;

      // Ensure that we do not overun our buffer.
      if(p_evt_write->len > BLE_MAX_RX_SVC_BYTES)
      {
        SEGGER_RTT_printf(0, "GATT message in %d bytes when max is %d\n", p_evt_write->len, BLE_MAX_RX_SVC_BYTES);
        break;
      }

      // Handle notification enable/disable p_evt_write->handle == p_nus->tx_handles.cccd_handle
      if(p_evt_write->handle == m_xtag_service.tx_char_handles.cccd_handle)
      {
        if(p_evt_write->data[0] == 0x01) { SEGGER_RTT_printf(0, "Notifications on tx enabled.\n"); }
        else { SEGGER_RTT_printf(0, "Notifications on tx disabled.\n"); }

        break;
      }

      // Get the incomming cmd bytes.
      memcpy(rx_char_data, p_evt_write->data, p_evt_write->len);

      // Schedule handling on main() via scheduler.
      app_sched_event_put(rx_char_data,  p_evt_write->len, on_main_shed_rx_char_evt_handler);

      break;
    }

    default:
      // No implementation needed.
      break;
  }
}

void ble_stack_init(void)
{
  ret_code_t err_code;

  err_code = nrf_sdh_enable_request();
  APP_ERROR_CHECK(err_code);

  // Configure the BLE stack using the default settings.
  uint32_t ram_start = 0;
  err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
  APP_ERROR_CHECK(err_code);

  // Enable BLE stack. To see required RAM config changes, enable RTT within nrf_sdh_ble.c 
  // (where nrf_sdh_ble_enable() is implemented). Uncomment #include "SEGGER_RTT.h" and logs in nrf_sdh_ble_enable(). 
  err_code = nrf_sdh_ble_enable(&ram_start);
  APP_ERROR_CHECK(err_code);

  // Register a handler for BLE events.
  NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);
}

// Gap params init from egs.
void gap_params_init(void)
{
  ret_code_t              err_code;
  ble_opt_t               ble_opt;
  ble_gap_addr_t cur_gap_addr;
  ble_gap_conn_params_t   gap_conn_params;
  ble_gap_conn_sec_mode_t sec_mode;
  uint8_t id[BLE_GAP_ADDR_LEN];

  // Add extended connection event support to send multiple packets per event.
  memset(&ble_opt, 0x00, sizeof(ble_opt));
  ble_opt.common_opt.conn_evt_ext.enable = 1;
  err_code = sd_ble_opt_set(BLE_COMMON_OPT_CONN_EVT_EXT, &ble_opt);
  APP_ERROR_CHECK(err_code);

  // From egs.
  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

  // Get the address from the device.
  err_code = sd_ble_gap_addr_get(&cur_gap_addr);
  APP_ERROR_CHECK(err_code);

  // Save off the MAC format as our ID (local and in xtag_data).
  memcpy(id, cur_gap_addr.addr, BLE_GAP_ADDR_LEN);

  // Save back ensuring type is public.
  cur_gap_addr.addr_type = BLE_GAP_ADDR_TYPE_PUBLIC;
  sd_ble_gap_addr_set(&cur_gap_addr);
  APP_ERROR_CHECK(err_code);

  // Setup and use a device name that includes the ID suffix bytes.
  char name_w_id[strlen(DEVICE_NAME) + DEVICE_NAME_SUFFIX_LEN];
  memset(name_w_id, 0, sizeof(name_w_id));
  sprintf(name_w_id, "%s %02X%02X%02X%02X%02X%02X", DEVICE_NAME, id[5], id[4], id[3], id[2], id[1], id[0]);
  err_code = sd_ble_gap_device_name_set(&sec_mode, (const uint8_t *)name_w_id, strlen(name_w_id));
  APP_ERROR_CHECK(err_code);

  // We will expose ourselves as a generic "tag".
  err_code = sd_ble_gap_appearance_set(BLE_APPEARANCE_GENERIC_TAG);
  APP_ERROR_CHECK(err_code);

  memset(&gap_conn_params, 0, sizeof(gap_conn_params));

  gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
  gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
  gap_conn_params.slave_latency     = SLAVE_LATENCY;
  gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

  err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
  APP_ERROR_CHECK(err_code);
}

// Function for handling events from the GATT MODULE. See ble for other GATT events.
static void gatt_evt_handler(nrf_ble_gatt_t * p_gatt, nrf_ble_gatt_evt_t const * p_evt)
{
  switch (p_evt->evt_id)
  {
    case NRF_BLE_GATT_EVT_ATT_MTU_UPDATED:

      SEGGER_RTT_printf(0, "ATT MTU updated from central after connection. MTU is %u bytes.\n", 
        p_evt->params.att_mtu_effective);

      break;

    case NRF_BLE_GATT_EVT_DATA_LENGTH_UPDATED:

      SEGGER_RTT_printf(0, "Data length updated from central after connection to %u bytes.\n",
        p_evt->params.data_length);

      break;

    default:  // Ignore the rest.

      break;
  }
}

/// Init GATT
void gatt_init(void)
{
  ret_code_t err_code = nrf_ble_gatt_init(&m_gatt, gatt_evt_handler);
  APP_ERROR_CHECK(err_code);
}

// Init services.
void services_init(ble_os_t * p_xtag_service)
{
  uint32_t      err_code;

  // Setup our custom service for comms.
  ble_uuid_t service_uuid;

  // Declare 16-bit service and 128-bit base UUIDs and add them to the BLE stack
  ble_uuid128_t base_uuid = BLE_UUID_XTAG_BASE_UUID;
  service_uuid.uuid = BLE_UUID_XTAG_SERVICE;
  err_code = sd_ble_uuid_vs_add(&base_uuid, &service_uuid.type);
  APP_ERROR_CHECK(err_code);

  // Setup the default handle as per egs.
  p_xtag_service->conn_handle = BLE_CONN_HANDLE_INVALID;

  // Add our service
  err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY,
                                      &service_uuid,
                                      &p_xtag_service->service_handle);
  APP_ERROR_CHECK(err_code);

  SEGGER_RTT_printf(0, "Executing our_service_init().\n"); // Print message to RTT to the application flow
  SEGGER_RTT_printf(0, "Service UUID: 0x%#04x\n", service_uuid.uuid); // Print service UUID should match definition BLE_UUID_OUR_SERVICE
  SEGGER_RTT_printf(0, "Service UUID type: 0x%#02x\n", service_uuid.type); // Print UUID type. Should match BLE_UUID_TYPE_VENDOR_BEGIN. Search for BLE_UUID_TYPES in ble_types.h for more info
  SEGGER_RTT_printf(0, "Service handle: 0x%#04x\n", p_xtag_service->service_handle); // Print out the service handle. Should match service handle shown in MCP under Attribute values

  // Add the RX Characteristic. This is taken directly from ble_nus.c to match NUS.
  ble_add_char_params_t add_char_params;
  memset(&add_char_params, 0, sizeof(add_char_params));
  add_char_params.uuid                     = BLE_UUID_XTAG_RX_CHARACTERISTC_UUID;
  add_char_params.uuid_type                = service_uuid.type;
  add_char_params.max_len                  = BLE_MAX_RX_SVC_BYTES;
  add_char_params.init_len                 = sizeof(uint8_t);
  add_char_params.is_var_len               = true;
  add_char_params.char_props.write         = 1;
  add_char_params.char_props.write_wo_resp = 1;

  add_char_params.read_access  = SEC_OPEN;
  add_char_params.write_access = SEC_OPEN;

  characteristic_add(p_xtag_service->service_handle, &add_char_params, &p_xtag_service->rx_char_handles);

  // Add the TX Characteristic. This is taken directly from ble_nus.c to match NUS.
  memset(&add_char_params, 0, sizeof(add_char_params));
  add_char_params.uuid              = BLE_UUID_XTAG_TX_CHARACTERISTC_UUID;
  add_char_params.uuid_type         = service_uuid.type;
  add_char_params.max_len           = BLE_MAX_TX_SVC_BYTES;
  add_char_params.init_len          = sizeof(uint8_t);
  add_char_params.is_var_len        = true;
  add_char_params.char_props.notify = 1;

  add_char_params.read_access       = SEC_OPEN;
  add_char_params.write_access      = SEC_OPEN;
  add_char_params.cccd_write_access = SEC_OPEN;

  characteristic_add(p_xtag_service->service_handle, &add_char_params, &p_xtag_service->tx_char_handles);

  // To support Dev Info Service Profile details.
  ble_dis_init_t dis_init;

  // Initialize Device Information Service.
  memset(&dis_init, 0, sizeof(dis_init));

  // Setup DIS attributes that we care about. Others null and won't be shown to clients.
  ble_srv_ascii_to_utf8(&dis_init.manufact_name_str, "Deviceworx");
  ble_srv_ascii_to_utf8(&dis_init.model_num_str, "xTAG-BLE-A");
  ble_srv_ascii_to_utf8(&dis_init.hw_rev_str, "1.00.01");
  ble_srv_ascii_to_utf8(&dis_init.fw_rev_str, "1.03.04");

  dis_init.dis_char_rd_sec = SEC_OPEN;

  err_code = ble_dis_init(&dis_init);
  APP_ERROR_CHECK(err_code);

  
}

// Connection parameters.
static void on_conn_params_evt(ble_conn_params_evt_t * p_evt)
{
  ret_code_t err_code;

  if (p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED)
  {
    err_code = sd_ble_gap_disconnect(m_xtag_service.conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
    APP_ERROR_CHECK(err_code);
  }
}

// Connection parameters error hander.
static void conn_params_error_handler(uint32_t nrf_error) { APP_ERROR_HANDLER(nrf_error); }

void conn_params_init(void)
{
  ret_code_t             err_code;
  ble_conn_params_init_t cp_init;

  memset(&cp_init, 0, sizeof(cp_init));

  cp_init.p_conn_params                  = NULL;
  cp_init.first_conn_params_update_delay = FIRST_CONN_PARAMS_UPDATE_DELAY;
  cp_init.next_conn_params_update_delay  = NEXT_CONN_PARAMS_UPDATE_DELAY;
  cp_init.max_conn_params_update_count   = MAX_CONN_PARAMS_UPDATE_COUNT;
  cp_init.start_on_notify_cccd_handle    = BLE_GATT_HANDLE_INVALID;
  cp_init.disconnect_on_fail             = false;
  cp_init.evt_handler                    = on_conn_params_evt;
  cp_init.error_handler                  = conn_params_error_handler;

  err_code = ble_conn_params_init(&cp_init);
  APP_ERROR_CHECK(err_code);
}

// Config advertising.
void advertising_config(void)
{
  ret_code_t             err_code;

  // Setup var to hold ble_advdata_t advert data. 
  ble_advdata_t advdata;
  memset(&advdata, 0, sizeof(advdata));

  // Setup var to hold ble_advdata_t scan resp data. 
  ble_advdata_t srdata;
  memset(&srdata, 0, sizeof(srdata));

  // Just use srdata to indicate that we want to expose the name provided in gap_params_init().
  srdata.name_type = BLE_ADVDATA_FULL_NAME;
  
  // Setup variable to hold manufacturer specific advert data
  ble_advdata_manuf_data_t mfr_advdata; 
  memset(&mfr_advdata, 0, sizeof(mfr_advdata));

  // Update mfr_data and add to advdata.
  uint8_t data[MFR_PAYLOAD_SIZE_GAP];
  for(int x=1; x<MFR_PAYLOAD_SIZE_GAP; x++) { data[x]=x; } // For nordic troubleshooting we don't care about adv content.

  mfr_advdata.company_identifier =  0x030E; //Dworx company ID
  mfr_advdata.data.p_data = data;
  mfr_advdata.data.size = sizeof(data);
  advdata.name_type = BLE_ADVDATA_NO_NAME;
  advdata.p_manuf_specific_data = &mfr_advdata;

  advdata.include_appearance      = true;
  advdata.flags                   = BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE;
  advdata.uuids_complete.uuid_cnt = sizeof(m_adv_uuids) / sizeof(m_adv_uuids[0]);
  advdata.uuids_complete.p_uuids  = m_adv_uuids;

  // Initialize advertising parameters (used when starting advertising).
  memset(&m_adv_params, 0, sizeof(m_adv_params));

  m_adv_params.properties.type = BLE_GAP_ADV_TYPE_CONNECTABLE_SCANNABLE_UNDIRECTED;
  m_adv_params.p_peer_addr     = NULL;    // Undirected advertisement.
  m_adv_params.filter_policy   = BLE_GAP_ADV_FP_ANY;
  m_adv_params.interval        = APP_ADV_INTERVAL;
  m_adv_params.duration        = 950;       // 10 mSec counts - 100 for each sec. Timer restarts every 10 sec, so use 9.5 sec.

  // Encode advdata to m_gapData[m_bufIdx]
  err_code = ble_advdata_encode(&advdata, m_gapData[m_bufIdx].adv_data.p_data, (uint16_t *) &m_gapData[m_bufIdx].adv_data.len);
  APP_ERROR_CHECK(err_code);

  // Encode srdata to m_gapData[m_bufIdx]
  err_code = ble_advdata_encode(&srdata, m_gapData[m_bufIdx].scan_rsp_data.p_data, (uint16_t *) &m_gapData[m_bufIdx].scan_rsp_data.len);
  APP_ERROR_CHECK(err_code);
 
  err_code = sd_ble_gap_adv_set_configure(&m_adv_handle, &m_gapData[m_bufIdx], &m_adv_params);
  APP_ERROR_CHECK(err_code);

  err_code = sd_ble_gap_tx_power_set(BLE_GAP_TX_POWER_ROLE_ADV,m_adv_handle, 4); 
  APP_ERROR_CHECK(err_code);
}

/**@brief Application main function.
 */
int main(void)
{
    bool erase_bonds;

    SEGGER_RTT_Init();
    SEGGER_RTT_printf(0, "Peripheral Logging started.\n");

    // As per egs, setup the clock driver to use the lfclk (before timers_init)
    APP_ERROR_CHECK(nrf_drv_clock_init());
    nrf_drv_clock_lfclk_request(NULL);
  
    // Additional Inits.
    APP_SCHED_INIT(SCHED_MAX_EVENT_DATA_SIZE, SCHED_QUEUE_SIZE);
    timers_init();
    APP_ERROR_CHECK(nrf_pwr_mgmt_init());
    APP_ERROR_CHECK(nrf_mem_init());

    // Alloc RAM.
    m_p_ramstore = (uint8_t*) nrf_malloc(240);

    // Setup a pattern of 0 - 239 in the store.
    for(int x=0; x<240; x++) { m_p_ramstore[x] = x; }

    // More inits.
    ble_stack_init();
    gap_params_init();
    gatt_init();
    services_init(&m_xtag_service);
    conn_params_init();

    advertising_config();

    // Start advertising..
    sd_ble_gap_adv_start(m_adv_handle, APP_BLE_CONN_CFG_TAG);

    // Start the 100 msec timer.
    APP_ERROR_CHECK(app_timer_start(m_100_msec_timer_id, APP_TIMER_TICKS(100), NULL));

    // Enable DCDC mode (low power support in idle mode).
    APP_ERROR_CHECK(sd_power_dcdc_mode_set(NRF_POWER_DCDC_ENABLE));

    SEGGER_RTT_printf(0, "BLE peripheral example started!");

    // Enter main loop.
    for (;;) 
    { 
      nrf_pwr_mgmt_run(); 

      // Handled scheduler actions.
      app_sched_execute();
    }
}


/**
 * @}
 */
